#include "GLview.hpp"

#include <iostream>
#include <algorithm>
#include <cmath>

using namespace std;

float copysign0(float x, float y) { return (y == 0.0f) ? 0 : copysign(x,y); }

void Matrix2Quaternion(QQuaternion &Q, QMatrix4x4 &M) {
  Q.setScalar(sqrt( max( 0.0f, 1 + M(0,0)  + M(1,1) + M(2,2) ) ) / 2);
  Q.setX(sqrt( max( 0.0f, 1 + M(0,0) - M(1,1) - M(2,2) ) ) / 2);
  Q.setY(sqrt( max( 0.0f, 1 - M(0,0) + M(1,1) - M(2,2) ) ) / 2);
  Q.setZ(sqrt( max( 0.0f, 1 - M(0,0) - M(1,1) + M(2,2) ) ) / 2);
  Q.setX(copysign0( Q.x(), M(2,1) - M(1,2) ));  Q.setY(copysign0( Q.y(), M(0,2) - M(2,0) ));  Q.setZ(copysign0( Q.z(), M(1,0) - M(0,1) )) ;
}

GLview::GLview(QWidget *parent)  : QOpenGLWidget(parent) {
  scaleFlag = translateFlag = rotateFlag = false;
  lastPosFlag = false;

  thetaz = thetay = thetax = 0;

  startTimer(20, Qt::PreciseTimer);
  elapsed_time.start(); elapsed_time.invalidate();  
  timeAccumulator = 0; totalTime = 0;
  
}  

// Use to read in geometry.
void readVectors(QVector<QVector3D> &result, const QString filename) {
  QFile file(filename);
  if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {  return; }

  QTextStream in(&file);
  while (!in.atEnd()) {
    QString line = in.readLine(); line = line.trimmed();
    QStringList tokens = line.split(' ', QString::SkipEmptyParts);
    if(tokens.size() == 3) {
      QVector3D v(tokens[0].toFloat(), tokens[1].toFloat(), tokens[2].toFloat());
      result.push_back(v);
    }
  }
}

void GLview::initializeGL() {
  initializeOpenGLFunctions();
  vao.create(); if (vao.isCreated()) vao.bind();

  glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );   // Set the clear color to black

  glEnable(GL_DEPTH_TEST);    // Enable depth buffer

  // Prepare a complete shader program...
  if ( !prepareShaderProgram(phong_shader,  ":/perfrag.vsh", ":/perfrag.fsh" ) ) return;

  num_vertices = 0;   // Load geometry.

  QVector<QVector3D> vertices, normals;
  readVectors(vertices, ":/geometry_vertices.txt"); readVectors(normals, ":/geometry_normals.txt");

  if(vertices.size() != normals.size() ) {
    qWarning() << "vertices.size() != normals.size()" << endl; exit(1);
  }
  num_vertices = vertices.size();

  vertexBuffer.create();
  vertexBuffer.setUsagePattern( QOpenGLBuffer::DynamicDraw );
  vertexBuffer.bind(); 
  vertexBuffer.allocate( &vertices[0], num_vertices * 3 * sizeof( float ) );

  normalBuffer.create();
  normalBuffer.setUsagePattern( QOpenGLBuffer::DynamicDraw );
  normalBuffer.bind();
  normalBuffer.allocate( &normals[0], num_vertices * 3 * sizeof( float ) );

  phong_shader.bind();

  vertexBuffer.bind(); 
  phong_shader.setAttributeBuffer( "VertexPosition", GL_FLOAT, 0, 3 );
  phong_shader.enableAttributeArray( "VertexPosition" );

  normalBuffer.bind();
  phong_shader.setAttributeBuffer( "VertexNormal", GL_FLOAT, 0, 3 );
  phong_shader.enableAttributeArray( "VertexNormal" );

  // Set default shading parameters. 
  LightDirection = QVector3D(0,0, 1);
  LightIntensity = QVector3D(1,1,1);
  Ns = 100; 
  Kd = QVector3D(0.8, 0.8, 0.8);
  Ks = QVector3D(0.95, 0.95, 0.95);
  Ka = QVector3D(0, 0, 0);

  phong_shader.setUniformValue("LightIntensity", LightIntensity);
  phong_shader.setUniformValue("Kd", Kd);
  phong_shader.setUniformValue("Ks", Ks);
  phong_shader.setUniformValue("Ka", Ka);
  phong_shader.setUniformValue("Shininess", Ns);

  // Initialize camera
  yfov = 55;
  neardist = 1; fardist = 1000;

  eye = QVector3D(-3,3,3); lookCenter = QVector3D(0,0,0); lookUp = QVector3D(0,0,1);

  QMatrix4x4 view; view.lookAt(eye, lookCenter, lookUp);
  Matrix2Quaternion(camrot, view);
}

  // Set the viewport to window dimensions
void GLview::resizeGL( int w, int h ) {  glViewport( 0, 0, w, qMax( h, 1 ) ); }

void GLview::initCameraGL() {
  model.setToIdentity(); view.setToIdentity(); projection.setToIdentity();

  QVector3D z_axis(0,0,1),y_axis(0,1,0), x_axis(1, 0, 0);

  QQuaternion q_yaw = QQuaternion::fromAxisAndAngle(z_axis, thetaz);
  QQuaternion q_pitch = QQuaternion::fromAxisAndAngle(y_axis, thetay);
  QQuaternion q_roll = QQuaternion::fromAxisAndAngle(x_axis, thetax);
  QQuaternion q_euler = q_yaw * q_pitch * q_roll;
  q_euler.normalize();

  QVector3D model_delta(0,0,0);
  model.rotate(q_euler); model.translate(model_delta);

  view.rotate(camrot); view.translate(-eye);

  projection.perspective(yfov, (float)width() / (float)height(), neardist, fardist);

  QMatrix4x4 model_view = view * model;
  QMatrix3x3 normal_matrix = model_view.normalMatrix();
  //QMatrix4x4 model_view = model;
  //QMatrix3x3 normal_matrix = model.normalMatrix();
 
  phong_shader.bind();
  phong_shader.setUniformValue("ModelViewMatrix", model_view);
  phong_shader.setUniformValue("NormalMatrix", normal_matrix);
  phong_shader.setUniformValue("MVP", projection * model_view);
  phong_shader.setUniformValue("LightDirection", view * QVector4D(LightDirection, 0));
}

void GLview::paintGL() {
  initCameraGL();

  // Clear the buffer with the current clearing color
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  phong_shader.bind(); // Draw stuff and run shaders

  /*vertexBuffer.bind(); 
  phong_shader.setAttributeBuffer( "VertexPosition", GL_FLOAT, 0, 3 );
  phong_shader.enableAttributeArray( "VertexPosition" );

  normalBuffer.bind();
  phong_shader.setAttributeBuffer( "VertexNormal", GL_FLOAT, 0, 3 );
  phong_shader.enableAttributeArray( "VertexNormal" );*/

  glDrawArrays( GL_TRIANGLES, 0, num_vertices );
}

void GLview::keyPressGL(QKeyEvent* e) {
  switch ( e->key() ) {
  case Qt::Key_Escape:
    QCoreApplication::instance()->quit();
    break;
  default:
    QOpenGLWidget::keyPressEvent( e );
    break;
    // Set rotate, scale, translate modes (for track pad mainly).
  case Qt::Key_R: toggleRotate();    break;
  case Qt::Key_S: toggleScale();     break;
  case Qt::Key_T: toggleTranslate();     break;
  }
}


bool GLview::prepareShaderProgram(QOpenGLShaderProgram &prep_shader, 
				  const QString &vertex_file, const QString &fragment_file) {
    // First we load and compile the vertex shader.
    bool result = prep_shader.addShaderFromSourceFile( QOpenGLShader::Vertex, vertex_file );
    if ( !result ) qWarning() << prep_shader.log();

    // ...now the fragment shader...
    result = prep_shader.addShaderFromSourceFile( QOpenGLShader::Fragment, fragment_file );
    if ( !result ) qWarning() << prep_shader.log();

    // ...and finally we link them to resolve any references.
    result = prep_shader.link();
    if ( !result ) {   qWarning() << "Could not link shader program:" << prep_shader.log(); exit(1);   }
    return result;
}

void GLview::mousePressEvent(QMouseEvent *event) { 
  float px = event->x(), py = event->y();
  float x = 2.0 * (px + 0.5) / float(width())  - 1.0,  y = -(2.0 * (py + 0.5) / float(height()) - 1.0); 
  lastPosX = x; lastPosY = y;
  lastPosFlag = true; 
  event->accept(); 
}

void GLview::mouseMoveEvent(QMouseEvent *event) {
  float px = event->x(), py = event->y();
  float x = 2.0 * (px + 0.5) / float(width())  - 1.0;
  float y = -(2.0 * (py + 0.5) / float(height()) - 1.0); 

  // Record a last position if none has been set.  
  if(!lastPosFlag) { lastPosX = x; lastPosY = y; lastPosFlag = true; return; }
  float dx = x - lastPosX, dy = y - lastPosY;
  lastPosX = x; lastPosY = y; // Remember mouse position.

  if (rotateFlag || (event->buttons() & Qt::LeftButton)) { // Rotate scene around a center point.   
    float theta_y = 2.0 * dy / M_PI * 180.0f;
    float theta_x = 2.0 * dx / M_PI * 180.0f;

    QQuaternion revQ = camrot.conjugate();
    QQuaternion newrot = QQuaternion::fromAxisAndAngle(lookUp, theta_x);
    revQ = newrot * revQ;

    QVector3D side = revQ.rotatedVector(QVector3D(1,0,0));
    QQuaternion newrot2 = QQuaternion::fromAxisAndAngle(side, theta_y);
    revQ = newrot2 * revQ;
    revQ.normalize();
    
    camrot = revQ.conjugate().normalized();    

    eye = newrot.rotatedVector(eye - lookCenter) + lookCenter;
    eye = newrot2.rotatedVector(eye - lookCenter) + lookCenter;
  } 

  if (scaleFlag || (event->buttons() & Qt::MidButton)) { // Scale the scene.
    float factor = dx + dy;
    factor = exp(2.0 * factor);
    factor = (factor - 1.0) / factor;
    QVector3D translation = (lookCenter - eye) * factor; 
    eye += translation;
  }
  if (translateFlag || (event->buttons() & Qt::RightButton)) { // Translate the scene.
    QQuaternion revQ = camrot.conjugate().normalized();
    QVector3D side = revQ.rotatedVector(QVector3D(1,0,0));
    QVector3D upVector = revQ.rotatedVector(QVector3D(0,1,0));

    float length = lookCenter.distanceToPoint(eye) * tanf(yfov * M_PI / 180.0f);
    QVector3D translation = -((side * (length * dx)) + (upVector * (length * dy) ));
    eye += translation;
    lookCenter += translation;
  }
  event->accept();
  //  update(); not needed since in timeout
}

void GLview::updateGLview(float dt) {
  thetaz += dt * 35;
  if(thetaz > 360) { thetaz = 0; }
}

void GLview::timerEvent(QTimerEvent *) {
  if(!elapsed_time.isValid()) {   elapsed_time.restart(); return;  } // Skip first udpate.
  qint64 nanoSec = elapsed_time.nsecsElapsed();
  elapsed_time.restart();  

  double dt = 0.01;
  double frameTime = double(nanoSec) * 1e-9;
  timeAccumulator += frameTime;
  while ( timeAccumulator >= dt ) {
    updateGLview(dt);  totalTime += dt;  timeAccumulator -= dt;
  }
  update();    
}
